@@ -24,6 +24,16 @@
 #else // Use dynamic library by default.
 #define TIMESHOP_CLASS Q_DECL_IMPORT
 #endif
+#ifndef WIN32
+#ifndef TIMESHOP_DATADIR
+#define TIMESHOP_DATADIR "/usr/share/timeshop"
+#endif // TIMESHOP_DATADIR
+#endif // WIN32
+#ifdef Q_WS_HILDON
+#define TIMESHOP_CREATE_PHONON_PATH
+#else
+#define TIMESHOP_PREPARE_ONCE
+#endif
 
 #include <utility>
 #include <QTimer>
@@ -36,25 +46,48 @@
 #include <QDateTime>
 #include <QAbstractListModel>
 #include <QTreeView>
+#include <QThread>
+#include <QMutex>
+// Choose alarm type
+#ifdef Q_WS_MAEMO_5
+#define TIMESHOP_USE_MAFW
+#else
 #ifdef QT_PHONON_LIB
 #define TIMESHOP_USE_PHONON
-#endif
-#ifdef TIMESHOP_USE_PHONON
+#endif // QT_PHONON_LIB
+#endif // QT_WS_MAEMO_5
+#if defined( TIMESHOP_USE_PHONON )
 #include <mediaobject.h>
-#else
+#include <audiooutput.h>
+#elif !defined( TIMESHOP_USE_MAFW )
 #include <QSound>
-#endif
+#endif // TIMESHOP_USE_PHONON
+
 #include <QXmlStreamWriter>
 #include <QFile>
 #include <QGridLayout>
 #ifdef Q_WS_MAEMO_5
 #define Q_WS_HILDON
 #endif
+// #define TIMESHOP_WIDGET_DISPLAY_DRAW_TEXT_RECT
+#ifdef Q_WS_HILDON
+#define TIMESHOP_MAEMO5_WORKAROUND_DIALOG
+// #define TIMESHOP_DEBUG_MAFW
+// #define TIMESHOP_USE_DBUS
+// #define TIMESHOP_USE_LIBPLAYBACK
+#define TIMESHOP_FINGER_FRIENDLY
+#endif // Q_WS_HILDON
+#if defined( TIMESHOP_USE_PHONON ) && ( defined( TIMESHOP_USE_MAFW ) || defined( TIMESHOP_USE_DBUS ) )
+#define TIMESHOP_PLAYER_CONTROL
+#endif
 
 //! \note Major refactoring is planned. StopwatchWidget and AlarmTimerWidget classes and their public slots and signals will stay as they are.
 class Ui_SettingsDialog;
 namespace Timeshop
 {
+  // These finctions must be called from the main()
+  void TIMESHOP_CLASS init();
+  void TIMESHOP_CLASS cleanup();
   //! \todo Move to utility library.
   //=== This classes will be moved to separate library.
   template <typename BitsEnum = char, typename StorageType = unsigned char> class Flags
@@ -116,12 +149,12 @@ namespace Timeshop
   }; // AboutAction
 
   //! \note Highly experimental. Don't use it directly!
-  class Work;
-  class Persistent
+  class TIMESHOP_CLASS Work;
+  class TIMESHOP_CLASS Persistent
   {
   public:
     typedef int ID;
-    class Loader
+    class TIMESHOP_CLASS Loader
     {
     public:
       typedef QList<Loader*> List;
@@ -179,21 +212,29 @@ namespace Timeshop
     // Object was loaded (maybe not successfully though). Overload for final touches.
     virtual bool object_loaded( Work* Root = 0 );
     virtual void write( QXmlStreamWriter& Stream ) const;
+    // Override to return true to save options to settings file
+    virtual bool has_options() const { return false; }
+    // This method saves options for this object to settings file. To do it override has_options and write_options methods.
+    void save_options( QXmlStreamWriter& Stream ) const;
   protected:
     virtual void write_attributes( QXmlStreamWriter& Stream ) const;
     virtual void write_elements( QXmlStreamWriter& Stream ) const;
+    virtual void write_options( QXmlStreamWriter& Stream ) const;
     ID ObjectID;
     static ID LastID;
   }; // Persistent
-  //=== Move out end
 
   typedef qint64 Time;
+  Time elapsed( const QDateTime& From, const QDateTime& To );
+  QString format_time( Time TimeValue, int Precision = 0 );
+  //=== Move out end
+
   // This will evolve into somthing like "TimeProvider in TimeSystem".
   class TIMESHOP_CLASS Timer : public QObject
   {
     Q_OBJECT
   public:
-    class Client
+    class TIMESHOP_CLASS Client
     {
     public:
       virtual ~Client() {}
@@ -212,7 +253,7 @@ namespace Timeshop
   {
   public:
     typedef Persistent Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -223,7 +264,7 @@ namespace Timeshop
     typedef Commands::Bit Command;
     typedef Flags<Command> CommandSet;
     // Target to receive commands from controller. Can anyone give better name?
-    class Client
+    class TIMESHOP_CLASS Client
     {
     public:
       virtual ~Client() {}
@@ -251,6 +292,7 @@ namespace Timeshop
     virtual void enable( CommandSet ToEnable, bool Enable = true );
     virtual void disable( CommandSet ToDisable ) { enable( ToDisable, false ); }
     // Show/hide command controls (if any).
+    virtual void swap( CommandSet ToHide, CommandSet ToShow ); //!< Swap buttons. It's better than hide and show, because can be optimized a lot.
     virtual void show( CommandSet ToShow, bool Show = true );
     virtual void hide( CommandSet ToHide ) { show( ToHide, false ); }
     // Don't change status (enable/disable/hide/show) of these commands.
@@ -272,7 +314,7 @@ namespace Timeshop
     Q_OBJECT
   public:
     typedef Controller Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -318,7 +360,7 @@ namespace Timeshop
     Q_OBJECT
   public:
     typedef WidgetController Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -332,6 +374,7 @@ namespace Timeshop
     // Enable/disable command button
     void enable( CommandSet ToEnable, bool Enable = true );
     // Show/hide command button
+    void swap( CommandSet ToHide, CommandSet ToShow );
     void show( CommandSet ToShow, bool Show = true );
     bool add( QPushButton* Button, Command ForCommand, Qt::Alignment Align = 0, int Stretch = 0 );
     const char* loader_tag() const { return Loader::Tag; }
@@ -357,13 +400,13 @@ namespace Timeshop
   {
   public:
     typedef Persistent Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
       Loader( const char* Tag0 = Tag ) : Super::Loader( Tag0 ) {}
     }; // Loader
-    class Mark
+    class TIMESHOP_CLASS Mark
     {
     public:
       typedef QList<Mark> List;
@@ -410,14 +453,87 @@ namespace Timeshop
     Controller::List Controllers; // Controllers from wich we receive commands.
     Time LastMark;
     Mark::List Marks;
+    bool FirstTick;
   }; // Stopwatch
 
+  // Abstract alarm
+  class TIMESHOP_CLASS Alarm
+  {
+  public:
+    enum PlayerControlMode { PlayerControlNone, PlayerControlPauseResume, PlayerControlStop };
+    Alarm() {}
+    virtual ~Alarm() {}
+    virtual void start() = 0;
+    virtual void stop() = 0;
+    virtual void reset() = 0;
+    virtual void change_sound( const QString& SoundFile ) = 0;
+#ifdef TIMESHOP_PLAYER_CONTROL
+    virtual void change_player_control( PlayerControlMode NewPlayerControl ) = 0;
+#endif // TIMESHOP_PLAYER_CONTROL
+  }; // Alarm
+
+#if defined( TIMESHOP_USE_PHONON )
+#ifdef TIMESHOP_USE_LIBPLAYBACK
+  class Playback;
+#endif
+  // Alarm sound through phonon subsystem
+  class TIMESHOP_CLASS PhononAlarm : public QObject, public Alarm
+  {
+    Q_OBJECT
+  public:
+    PhononAlarm( const QString& SoundFile	
+#ifdef TIMESHOP_PLAYER_CONTROL
+		 , PlayerControlMode PlayerControl0 = Alarm::PlayerControlPauseResume
+#endif // TIMESHOP_PLAYER_CONTROL
+		 );
+    ~PhononAlarm();
+    void init_player( const QString& SoundFile );
+    void start();
+    void stop();
+    void reset();
+    void change_sound( const QString& SoundFile );
+#ifdef TIMESHOP_PLAYER_CONTROL
+    void change_player_control( PlayerControlMode NewPlayerControl ) { PlayerControl = NewPlayerControl; }
+#endif // TIMESHOP_PLAYER_CONTROL
+  public slots:
+    void state_changed( Phonon::State NewState, Phonon::State OldState );
+  protected:
+#ifdef TIMESHOP_PLAYER_CONTROL
+    void pause_player(); // Stop if we prefer it.
+    void resume_player();
+    bool ResumePlayback;
+    PlayerControlMode PlayerControl;
+#ifdef TIMESHOP_PLAYER_CONTROL_RESTORE_POSITION
+    qint Position;
+#endif // TIMESHOP_PLAYER_CONTROL_RESTORE_POSITION
+#endif // TIMESHOP_PLAYER_CONTROL
+#ifdef TIMESHOP_USE_LIBPLAYBACK
+    Playback* PlaybackObject;
+#endif
+    Phonon::MediaObject* AlarmSound;
+  }; // PhononAlarm
+#elif !defined( TIMESHOP_USE_MAFW ) // MAFWAlarm header includes only into sources that need it.
+  // Alarm sound through standard system (if you don't have Phonon). Deprecated because QSound is broken on some platforms.
+  class TIMESHOP_CLASS QSoundAlarm : public Alarm
+  {
+  public:
+    QSoundAlarm( const QString& SoundFile );
+    ~QSoundAlarm();
+    void start();
+    void stop();
+    void reset();
+    void change_sound( const QString& SoundFile );
+  protected:
+    QSound* AlarmSound;
+  }; // QSoundAlarm
+#endif // TIMESHOP_USE_PHONON
+
   // Archaic digital clock display
   class TIMESHOP_CLASS LCDDisplay : public QLCDNumber, public TimeDisplay
   {
   public:
     typedef TimeDisplay Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -440,7 +556,7 @@ namespace Timeshop
   {
   public:
     typedef TimeDisplay Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -456,8 +572,10 @@ namespace Timeshop
     void write_attributes( QXmlStreamWriter& Stream ) const;
     void write_elements( QXmlStreamWriter& Stream ) const;
     void resize_text( const QString& String );
+    void adjust_text_position( const QString& String );
     void paintEvent( QPaintEvent* Event );
     void resizeEvent( QResizeEvent* Event );
+    int Padding; // Padding from the right side to compensate wrong info about text sizes.
     QPoint TextPlace;
     QString TimeString;
   }; // WidgetDisplay
@@ -467,7 +585,7 @@ namespace Timeshop
   {
   public:
     typedef TimeDisplay Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -487,27 +605,34 @@ namespace Timeshop
     QString Text;
   }; // TitleDisplay
 
-  class MarksListWidget : public QTreeView
+  class TIMESHOP_CLASS MarksListModel : public QAbstractListModel
+  {
+    Q_OBJECT
+  public:
+    enum ColumnNum { NumberCol = 0, SplitCol, LapCol, CommentCol };
+    MarksListModel( Stopwatch::Mark::List& Marks0, int Precision0 = 0, QObject* Parent = 0 ) : QAbstractListModel( Parent ), Marks( Marks0 ), Precision( Precision0 ) {}
+    int rowCount( const QModelIndex& Parent = QModelIndex() ) const { return( Parent.isValid() ? 0 : Marks.size() ); }
+    int columnCount( const QModelIndex& Parent = QModelIndex() ) const { return( Parent.isValid() ? 0 : 4 ); }
+    QVariant data( const QModelIndex& Index, int Role = Qt::DisplayRole ) const;
+#ifndef TIMESHOP_MAEMO5_WORKAROUND_DIALOG
+    Qt::ItemFlags flags( const QModelIndex& Index ) const;
+    bool setData( const QModelIndex& Index, const QVariant& Value, int Role = Qt::EditRole );
+#endif // TIMESHOP_MAEMO5_WORKAROUND_DIALOG
+    QVariant headerData( int Section, Qt::Orientation Orientation, int Role = Qt::DisplayRole ) const;
+    Stopwatch::Mark* mark( int Index ) const;
+    int precision() const { return Precision; }
+    void precision( int NewPrecision ) { Precision = NewPrecision; } //!< \todo Update list items if precision has been changed.
+#ifdef TIMESHOP_MAEMO5_WORKAROUND_DIALOG
+  public slots:
+    void edit_mark( const QModelIndex& Index );
+#endif // TIMESHOP_MAEMO5_WORKAROUND_DIALOG
+  protected:
+    Stopwatch::Mark::List& Marks;
+    int Precision;
+  }; // MarksListModel
+  class TIMESHOP_CLASS MarksListWidget : public QTreeView
   {
   public:
-    class MarksListModel : public QAbstractListModel
-    {
-    public:
-      enum ColumnNum { NumberCol = 0, SplitCol, LapCol, CommentCol };
-      MarksListModel( Stopwatch::Mark::List& Marks0, int Precision0 = 0, QObject* Parent = 0 ) : QAbstractListModel( Parent ), Marks( Marks0 ), Precision( Precision0 ) {}
-      int rowCount( const QModelIndex& Parent = QModelIndex() ) const { return( Parent.isValid() ? 0 : Marks.size() ); }
-      int columnCount( const QModelIndex& Parent = QModelIndex() ) const { return( Parent.isValid() ? 0 : 4 ); }
-      QVariant data( const QModelIndex& Index, int Role = Qt::DisplayRole ) const;
-      Qt::ItemFlags flags( const QModelIndex& Index ) const;
-      bool setData( const QModelIndex& Index, const QVariant& Value, int Role = Qt::EditRole );
-      QVariant headerData( int Section, Qt::Orientation Orientation, int Role = Qt::DisplayRole ) const;
-      Stopwatch::Mark* mark( int Index ) const;
-      int precision() const { return Precision; }
-      void precision( int NewPrecision ) { Precision = NewPrecision; } //!< \todo Update list items if precision has been changed.
-    protected:
-      Stopwatch::Mark::List& Marks;
-      int Precision;
-    }; // MarksListModel
     MarksListWidget( Stopwatch::Mark::List& Marks0, int Precision0 = 0, QWidget* Parent = 0 );
     void appended(); // Inform that a mark was appended to the list
     int precision() const { return Model.precision(); }
@@ -528,7 +653,7 @@ namespace Timeshop
     int MarksListRow;
   public:
     typedef Stopwatch Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -554,6 +679,9 @@ namespace Timeshop
   public slots:
     virtual void show_marks( bool Show = true );
     virtual void save_marks();
+    virtual void save_marks_to_html( QIODevice& File );
+    virtual void save_marks_to_csv( QIODevice& File );
+    virtual void save_marks_to_xml( QIODevice& File );
    }; // StopwatchWidget
 
   //! \note Will be extended and maybe changed in the near future.
@@ -563,7 +691,7 @@ namespace Timeshop
     typedef Persistent Super;
     struct Period { enum { Millisecond = -3, Tenth = -1, Second = 0, Minute, Hour, Day, Week, Month, Year }; }; // Other negative values are valid for corresponding fractions.
     typedef QList<TimerSettings*> List;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -575,6 +703,14 @@ namespace Timeshop
     Time AlarmTime;
     int Precision;
     QString AlarmSound; //! \todo Make more flexible alarm with off/snooze button.
+    QStringList RecentSounds;
+    int MaxRecentSounds;
+    QString Layout;
+    QStringList RecentLayouts;
+    int MaxRecentLayouts;
+#ifdef TIMESHOP_PLAYER_CONTROL
+    Alarm::PlayerControlMode PlayerControl;
+#endif // TIMESHOP_PLAYER_CONTROL
   public:
     TimerSettings( Time InitTime0 = 0, int Precision0 = Period::Second, Time AlarmTime0 = 0, Persistent::ID ObjectID0 = -1 );
     virtual ~TimerSettings() {}
@@ -588,6 +724,16 @@ namespace Timeshop
     virtual void precision( int NewPrecision ) { Precision = NewPrecision; }
     const QString& alarm_sound() const { return AlarmSound; }
     virtual void alarm_sound( const QString& NewAlarmSound ) { AlarmSound = NewAlarmSound; }
+    void recent_sound( const QString& NewSound );
+    const QStringList& recent_sounds() const { return RecentSounds; }
+#ifdef TIMESHOP_PLAYER_CONTROL
+    Alarm::PlayerControlMode player_control() const { return PlayerControl; }
+    void player_control( Alarm::PlayerControlMode NewMode ) { PlayerControl = NewMode; }
+#endif // TIMESHOP_PLAYER_CONTROL
+    const QString& layout() const { return Layout; }
+    virtual void layout( const QString& NewLayout ) { Layout = NewLayout; }
+    void recent_layout( const QString& NewLayout );
+    const QStringList& recent_layouts() const { return RecentLayouts; }
     const char* loader_tag() const { return Loader::Tag; }
   protected:
     bool load_element( QXmlStreamReader& Stream, Work* Root = 0 );
@@ -616,8 +762,12 @@ namespace Timeshop
     SettingsDialog( TimerSettings& Settings0, QWidget* Parent = 0 );
     ~SettingsDialog();
   public slots:
-    void accept();
+#ifdef TIMESHOP_FINGER_FRIENDLY
+    void select_time();
+#endif
     void select_file();
+    void select_layout();
+    void accept();
   }; // SettingsDialog
 
   class TIMESHOP_CLASS TimerPreset : public TimerSettings
@@ -625,7 +775,7 @@ namespace Timeshop
     TimerPreset( const TimerSettings& ) {} // Disable this.
   public:
     typedef TimerSettings Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
@@ -634,7 +784,7 @@ namespace Timeshop
     }; // Loader
     struct Flags { enum Bit { InitTime = 0, AlarmTime, Precision, AlarmSound }; }; // In-class namespace Flags
     typedef QList<TimerPreset*> List;
-    class ListModel : public QAbstractListModel
+    class TIMESHOP_CLASS ListModel : public QAbstractListModel
     {
       List& Presets;
     public:
@@ -701,7 +851,7 @@ namespace Timeshop
   public:
     typedef StopwatchWidget Super;
     typedef QList<AlarmTimerWidget*> List;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static TimerSettings* find_settings( QXmlStreamReader& Stream, Work* Root );
@@ -711,6 +861,7 @@ namespace Timeshop
     }; // Loader
 
     AlarmTimerWidget( TimerSettings& Settings0, QWidget* Parent = 0, Persistent::ID ObjectID0 = -1 );
+    ~AlarmTimerWidget();
     bool create_default_layout();
     // Accessors
     TimerSettings& settings() const { return *Settings; }
@@ -727,8 +878,8 @@ namespace Timeshop
     Time time_count( const QDateTime& CurTime = QDateTime::currentDateTime()  );
     bool presets_visible() const { return ButtonsPad && ButtonsPad->isVisible(); }
     const char* loader_tag() const { return Loader::Tag; }
-  protected:
     bool load_element( QXmlStreamReader& Stream, Work* Root = 0 );
+  protected:
     void write_attributes( QXmlStreamWriter& Stream ) const;
     void write_elements( QXmlStreamWriter& Stream ) const;
   public slots:
@@ -745,29 +896,32 @@ namespace Timeshop
   protected:
     TimerSettings* Settings; // We don't delete this.
     bool Alarmed; // Alarm has signalled already.
-#ifdef TIMESHOP_USE_PHONON
-    Phonon::MediaObject* AlarmSound;
-#else
-    QSound* AlarmSound;
-#endif
+    Alarm* AlarmSound;
     QList<TimerPreset*> Presets;
     QWidget* ButtonsPad;
   }; // AlarmTimerWidget
 
   //! \note Experimental: This can be changed in 0.4.0.
-  class Work : public QObject, public Persistent
+  class TIMESHOP_CLASS Work : public QObject, public Persistent
   {
     Q_OBJECT
   public:
     typedef Persistent Super;
-    class Loader : public Super::Loader
+    class TIMESHOP_CLASS Loader : public Super::Loader
     {
     public:
       static const char Tag[];
       Loader( const char* Tag0 = Tag ) : Super::Loader( Tag0 ) {}
+#ifdef USE_WORK_LOADER_LOAD
       Persistent* load( QXmlStreamReader& Stream, Work* Root = 0 ) const;
+#else
+      Persistent* create_object( QXmlStreamReader& Stream, Work* Root, Persistent::ID ObjectID = -1 ) const;
+#endif
     }; // Loader
-    Work();
+    // Common & user timeshop data directories.
+    static QString data_dir();
+    static QString user_data_dir();
+    Work( Persistent::ID ObjectID = -1 );
     ~Work();
     const AlarmTimerWidget::List& timers() const { return Timers; }
     Stopwatch* find_timer( Persistent::ID ByID ) const;
@@ -780,11 +934,14 @@ namespace Timeshop
     bool add( TimerPreset& Preset );
     TimerPreset* preset( Persistent::ID ID );
     void add_timer( AlarmTimerWidget& NewTimer ); //!< \todo Make this a list Stopwatches, not ...Widgets.
-    Persistent::Loader* loader( const QString& Token );
+    AlarmTimerWidget* timer( Persistent::ID ID );
     QWidget* find_widget( const QString& ObjectName );
+    Persistent::Loader* loader( const QString& Token );
+  protected:
+    bool load_element( QXmlStreamReader& Stream, Work* Root = 0 );
   public slots:
     //! Remove timer from Work (on object destroy)
-    void remove_timer( QObject* NewTimer );
+    void remove_timer( QObject* OldTimer );
     // Overload to load/save this without new slots creation.
     virtual void load_from_stream() {}
     virtual void save_to_stream() {}
@@ -795,7 +952,7 @@ namespace Timeshop
     AlarmTimerWidget::List Timers;
   }; // Work
 
-  class XMLStreamReader : public QXmlStreamReader
+  class TIMESHOP_CLASS XMLStreamReader : public QXmlStreamReader
   {
   public:
     XMLStreamReader() {}
@@ -803,7 +960,7 @@ namespace Timeshop
     bool read_header();
   }; // XMLStreamReader
 
-  class XMLStreamWriter : public QXmlStreamWriter
+  class TIMESHOP_CLASS XMLStreamWriter : public QXmlStreamWriter
   {
   public:
     XMLStreamWriter() {}
@@ -814,7 +971,7 @@ namespace Timeshop
     void start();
     void close();
   }; // XMLStreamWriter
-  class XMLFileWriter : public XMLStreamWriter
+  class TIMESHOP_CLASS XMLFileWriter : public XMLStreamWriter
   {
     QFile File;
   public:
